@@ -103,6 +103,7 @@ use openpgp::armor;
 use openpgp::Cert;
 use openpgp::cert::prelude::*;
 use openpgp::Fingerprint;
+use openpgp::KeyID;
 use openpgp::packet::key::{
     PublicParts,
 };
@@ -126,6 +127,8 @@ use openpgp::policy::{
 use openpgp::serialize::SerializeInto;
 use openpgp::types::RevocationStatus;
 
+use openpgp::parse::buffered_reader;
+
 #[macro_use] mod log;
 #[macro_use] mod ffi;
 #[macro_use] pub mod rpm;
@@ -141,7 +144,9 @@ pub mod digest;
 lazy_static::lazy_static! {
     static ref P: RwLock<StandardPolicy<'static>> = RwLock::new(StandardPolicy::new());
 }
-const NP: &NullPolicy = &NullPolicy::new();
+const NP: &NullPolicy = unsafe {
+    &NullPolicy::new()
+};
 
 // Set according to the RPM_TRACE environment variable (enabled if
 // non-zero), or if we are built in debug mode.
@@ -437,10 +442,10 @@ fn _pgpDigParamsAlgo(dig: *const PgpDigParams,
     match (algotype, &dig.obj) {
         // pubkey algo.
         (PGPVAL_PUBKEYALGO, PgpDigParamsObj::Cert(cert)) => {
-            Ok(u8::from(cert.primary_key().pk_algo()).into())
+            Ok(u8::from(cert.primary_key().key().pk_algo()).into())
         }
         (PGPVAL_PUBKEYALGO, PgpDigParamsObj::Subkey(_, _)) => {
-            Ok(u8::from(dig.key().expect("valid").pk_algo()).into())
+            Ok(u8::from(dig.key().expect("valid").key().pk_algo()).into())
         }
         (PGPVAL_PUBKEYALGO, PgpDigParamsObj::Signature(sig)) => {
             Ok(u8::from(sig.pk_algo()).into())
@@ -539,10 +544,10 @@ fn _pgpDigParamsVersion(dig: *const PgpDigParams) -> c_int[0] {
     let dig = check_ptr!(dig);
     let version = match &dig.obj {
         PgpDigParamsObj::Cert(cert) => {
-            cert.primary_key().version()
+            cert.primary_key().key().version()
         }
         PgpDigParamsObj::Subkey(_, _) => {
-            dig.key().unwrap().version()
+            dig.key().unwrap().key().version()
         }
         PgpDigParamsObj::Signature(sig) => {
             sig.version()
@@ -566,13 +571,14 @@ fn _pgpDigParamsCreationTime(dig: *const PgpDigParams) -> u32[0] {
     let dig = check_ptr!(dig);
     let t = match &dig.obj {
         PgpDigParamsObj::Cert(cert) => {
-            cert.primary_key().creation_time()
+            cert.primary_key().key().creation_time()
         }
         PgpDigParamsObj::Subkey(cert, fpr) => {
             cert.keys().subkeys()
                 .key_handle(fpr)
                 .next()
                 .expect("subkey missing")
+                .key()
                 .creation_time()
         }
         PgpDigParamsObj::Signature(sig) => {
@@ -584,6 +590,43 @@ fn _pgpDigParamsCreationTime(dig: *const PgpDigParams) -> u32[0] {
        .as_secs() as u32)
 });
 
+ffi!(
+/// Returns a signature's salt.
+///
+/// Version 6 signatures are salted.  The salt needs to be fed into
+/// the digest buffer before the actual message, which is processed
+/// already by RPM.
+///
+/// The caller must *not* free the returned buffer.
+///
+/// Returns an error if the signature does not include a salt.
+/// Version 6 signatures always have a salt; version 3 and version 4
+/// signatures never have a salt and thus will always return an error.
+fn _pgpDigParamsSalt(dig: *const PgpDigParams,
+                     datap: *mut *const u8,
+                     lenp: *mut size_t)
+    -> ErrorCode
+{
+    let dig = check_ptr!(dig);
+    let datap = check_mut!(datap);
+    let lenp = check_mut!(lenp);
+
+    let sig = dig.signature().ok_or_else(|| {
+        Error::Fail("sig parameter does not designate a signature".into())
+    })?;
+
+    if let Some(salt) = sig.salt() {
+        t!("Salt: {}",
+           salt.iter().map(|v| format!("{:02X}", v)).collect::<String>());
+        *lenp = salt.len() as size_t;
+        *datap = salt.as_ptr();
+        Ok(())
+    } else {
+        Err(Error::Fail("The provided signature does not have any salt".into()))
+    }
+});
+
+
 ffi!(
 /// Verifies the signature.
 ///
@@ -662,7 +705,7 @@ fn _pgpVerifySignature2(key: *const PgpDigParams,
                     .and_then(|cert| {
                         cert.userids().next()
                             .map(|userid| {
-                                String::from_utf8_lossy(userid.value()).into_owned()
+                                String::from_utf8_lossy(userid.userid().value()).into_owned()
                             })
                     })
                     .unwrap_or_else(|| {
@@ -807,7 +850,7 @@ fn pgp_verify_signature(key: Option<&PgpDigParams>,
         let cert = key.cert().ok_or_else(|| {
             Error::Fail("key parameter is not a cert".into())
         })?;
-        let subkey = key.key().expect("is a certificate").fingerprint();
+        let subkey = key.key().expect("is a certificate").key().fingerprint();
 
         t!("Checking signature {} using {} with {} / {}",
            sig_id(), sig.hash_algo(),
@@ -873,31 +916,31 @@ fn pgp_verify_signature(key: Option<&PgpDigParams>,
         // Find the key.
         match vc.keys().key_handle(issuer.clone()).next() {
             Some(ka) => {
-                if ka.fingerprint() != subkey {
+                if ka.key().fingerprint() != subkey {
                     return_err!(None,
                                 "Key {} invalid: wrong subkey ({})",
-                                ka.keyid(), subkey);
+                                ka.key().keyid(), subkey);
                 }
                 if ! ka.for_signing() {
                     return_err!(None,
                                 "Key {} invalid: not signing capable",
-                                ka.keyid());
+                                ka.key().keyid());
                 }
                 if let Err(err) = ka.alive() {
                     legacy = true;
                     add_lint!(Some(err),
                               "Key {} invalid: key is not alive",
-                              ka.keyid());
+                              ka.key().keyid());
                 }
                 if let RevocationStatus::Revoked(_) = ka.revocation_status() {
                     legacy = true;
                     add_lint!(None,
                               "Key {} is invalid: key is revoked",
-                              ka.keyid());
+                              ka.key().keyid());
                 }
 
                 // Finally we can verify the signature.
-                sig.clone().verify_hash(&ka, ctx.ctx.clone())?;
+                sig.clone().verify_hash(ka.key(), ctx.ctx.clone())?;
                 if legacy {
                     return Err(Error::NotTrusted(
                         "Verification relies on legacy crypto".into())
@@ -927,13 +970,18 @@ fn pgp_verify_signature(key: Option<&PgpDigParams>,
 
         // Hash the signature into the context.
         match sig.version() {
-            4 => {
+            4 | 6 => {
                 sig_data.push(sig.version());
                 sig_data.push(sig.typ().into());
                 sig_data.push(sig.pk_algo().into());
                 sig_data.push(sig.hash_algo().into());
 
-                let l = sig.hashed_area().serialized_len() as u16;
+                let l = sig.hashed_area().serialized_len();
+                if sig.version() == 6 {
+                    // The v6 signatures encode the hashed length as 4 bytes
+                    sig_data.push((l >> 24) as u8);
+                    sig_data.push((l >> 16) as u8);
+                }
                 sig_data.push((l >> 8) as u8);
                 sig_data.push((l >> 0) as u8);
 
@@ -1257,6 +1305,15 @@ fn _pgpPubKeyCertLen(pkts: *const u8, pktslen: size_t,
                    t, start_of_packet);
                 break None;
             }
+
+            t => if t.is_critical() {
+                t!("Encountered a ({:?}) at offset {}, \
+                    which does not belong in a certificate",
+                   t, start_of_packet);
+                break None;
+            } else {
+                // Ignore unknown non-critical packet.
+            },
         }
 
         // Advance to the next packet.
@@ -1420,7 +1477,7 @@ fn pgp_prt_params(pkts: *const u8, pktlen: size_t,
                  // Unfortunately, the API only allows us to return
                  // one.
                  sig.get_issuers().into_iter().next()
-                     .map(|i| i.as_bytes().to_vec()),
+                     .map(|i| KeyID::from(&i)),
                  None)
             }
             Packet::PublicKey(_) | Packet::SecretKey(_)
@@ -1439,7 +1496,7 @@ fn pgp_prt_params(pkts: *const u8, pktlen: size_t,
                         "Failed to read an OpenPGP certificate"),
                 };
 
-                let keyid = cert.keyid().as_bytes().to_vec();
+                let keyid = cert.keyid().clone();
 
                 let userid = if let Ok(vc)
                     = cert.with_policy(&*P.read().unwrap(), None)
@@ -1447,7 +1504,7 @@ fn pgp_prt_params(pkts: *const u8, pktlen: size_t,
                     vc.primary_userid()
                         .ok()
                         .and_then(|u| {
-                            CString::new(u.value()).ok()
+                            CString::new(u.userid().value()).ok()
                         })
                 } else {
                     None
@@ -1485,15 +1542,7 @@ fn pgp_prt_params(pkts: *const u8, pktlen: size_t,
 
     let mut buffer: [u8; 8] = [0; 8];
     if let Some(issuer) = issuer {
-        let issuer = if issuer.len() > buffer.len() {
-            // We've got a fingerprint.  For v4 keys, the last 16
-            // bytes is the key id.
-            &issuer[issuer.len() - buffer.len()..]
-        } else {
-            &issuer[..]
-        };
-
-        for (i, c) in issuer.into_iter().enumerate() {
+        for (i, c) in issuer.as_bytes().into_iter().enumerate() {
             buffer[i] = *c as u8;
         }
     }
@@ -1545,7 +1594,7 @@ fn _pgpPrtParamsSubkeys(pkts: *const u8, pktlen: size_t,
         vc.primary_userid()
             .ok()
             .and_then(|u| {
-                CString::new(u.value()).ok()
+                CString::new(u.userid().value()).ok()
             })
     } else {
         None
@@ -1556,15 +1605,15 @@ fn _pgpPrtParamsSubkeys(pkts: *const u8, pktlen: size_t,
     let mut keys: Vec<*mut PgpDigParams> = cert
         .keys().subkeys()
         .map(|ka| {
-            t!("Subkey: {}", ka.keyid());
+            t!("Subkey: {}", ka.key().keyid());
 
             let zeros = [0; 8];
             let mut dig = PgpDigParams {
-                obj: PgpDigParamsObj::Subkey(cert.clone(), ka.fingerprint()),
+                obj: PgpDigParamsObj::Subkey(cert.clone(), ka.key().fingerprint()),
                 signid: zeros,
                 userid: userid.clone(),
             };
-            dig.signid.copy_from_slice(ka.keyid().as_bytes());
+            dig.signid.copy_from_slice(ka.key().keyid().as_bytes());
             move_to_c!(dig)
         })
         .collect();
@@ -1660,6 +1709,7 @@ fn _pgpPubKeyLint(pkts: *const c_char,
         lints.push(l.into());
     };
 
+    #[allow(clippy::never_loop)]
     let usable = 'done : loop {
         match cert.with_policy(&*P.read().unwrap(), None) {
             Err(err) => {
@@ -1705,7 +1755,7 @@ fn _pgpPubKeyLint(pkts: *const c_char,
 
         let mut have_signing = false;
         for ka in cert.keys() {
-            let keyid = ka.keyid();
+            let keyid = ka.key().keyid();
 
             match ka.with_policy(&*P.read().unwrap(), None) {
                 Err(err) => {
@@ -1759,11 +1809,11 @@ fn _pgpPubKeyLint(pkts: *const c_char,
                         continue;
                     }
 
-                    if ! ka.pk_algo().is_supported() {
+                    if ! ka.key().pk_algo().is_supported() {
                         lint(&format!("Subkey {} is not supported \
                                        (no support for {})",
                                       keyid,
-                                      ka.pk_algo()));
+                                      ka.key().pk_algo()));
                         continue;
                     }
 
@@ -1865,6 +1915,7 @@ fn dump_packets(pkts: &[u8]) -> Result<()> {
 
     fn symalgo(a: SymmetricAlgorithm) -> &'static str {
         use SymmetricAlgorithm::*;
+        #[allow(deprecated)]
         match a {
             Unencrypted => "Plaintext",
             IDEA => "IDEA",
@@ -1964,7 +2015,7 @@ fn dump_packets(pkts: &[u8]) -> Result<()> {
                         algos.iter()
                             .map(|a| {
                                 format!("{}({})",
-                                        a.to_string(),
+                                        a,
                                         Into::<u8>::into(*a))
                             })
                             .collect::<Vec<String>>()
@@ -2030,6 +2081,7 @@ fn dump_packets(pkts: &[u8]) -> Result<()> {
 
         // We only dump what rpm's internal OpenPGP implementation
         // dumps.  Other packets we silently ignore.
+        #[allow(deprecated)]
         match packet {
             Packet::Signature(sig) => {
                 // V4 Signature(2) DSA(17) SHA512(10) Generic certification of a User ID and Public Key(16)
@@ -2040,7 +2092,7 @@ fn dump_packets(pkts: &[u8]) -> Result<()> {
                           sig.version(),
                           pk_algo(sig.pk_algo()),
                           Into::<u8>::into(sig.pk_algo()),
-                          sig.hash_algo().to_string(),
+                          sig.hash_algo(),
                           Into::<u8>::into(sig.hash_algo()),
                           sigtype(sig.typ()),
                           Into::<u8>::into(sig.typ()));
@@ -2107,7 +2159,6 @@ fn dump_packets(pkts: &[u8]) -> Result<()> {
             Packet::SKESK(_skesk) => (),
             Packet::SEIP(_seip) => (),
             Packet::MDC(_mdc) => (),
-            Packet::AED(_aed) => (),
             _ => (),
         }
     }
@@ -2284,7 +2335,7 @@ fn _pgpPubkeyMerge(
     }
 
     let (merged, updated)
-	= cert1.clone().insert_packets2(cert2.into_packets2())?;
+	= cert1.clone().insert_packets(cert2.into_packets())?;
 
     let merged_bytes_;
     let (result, result_len) = if ! updated {
